/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/module/handler.c
 * @brief Handlers for incoming AFS requests (p2p and CS).
 * @author Christian Grothoff
 **/

#include "afs.h"
#include "bloomfilter.h"
#include "fileindex.h"
#include "manager.h"
#include "routing.h"
#include "policy.h"
#include "routing.h"

/* ********************* p2p handlers ****************** */

static int stat_p2p_query_count;
static int stat_p2p_chk_replies;
static int stat_p2p_3hash_replies;
static int stat_cs_query_count;
static int stat_cs_insert_chk_count;
static int stat_cs_insert_3hash_count;
static int stat_cs_index_block_count;
static int stat_cs_index_file_count;
static int stat_cs_index_super_count;
static int stat_cs_delete_chk_count;
static int stat_cs_delete_3hash_count;
static int stat_cs_unindex_block_count;
static int stat_cs_unindex_file_count;
static int stat_cs_unindex_super_count;

static int stat_cs_insert_sblock_count;
static int stat_cs_nsquery_count;
static int stat_p2p_nsquery_count;
static int stat_p2p_sblock_replies;


#define DEBUG_HANDLER NO

/**
 * Initialize the handler module. Registers counters
 * with the statistics module.
 *
 * @return OK on success, SYSERR on failure
 **/
int initAFSHandler() {
  stat_p2p_query_count 
    = statHandle("# p2p queries received");
  stat_p2p_chk_replies 
    = statHandle("# p2p CHK content received (kb)");
  stat_p2p_3hash_replies 
    = statHandle("# p2p search results received (kb)");
  stat_cs_query_count 
    = statHandle("# client queries received");
  stat_cs_insert_chk_count 
    = statHandle("# client CHK content inserted (kb)");
  stat_cs_insert_3hash_count 
    = statHandle("# client 3HASH search results inserted (kb)");
  stat_cs_index_block_count 
    = statHandle("# client file index requests received");
  stat_cs_index_file_count 
    = statHandle("# file index requests received");
  stat_cs_index_super_count 
    = statHandle("# super query index requests received");
  stat_cs_delete_chk_count 
    = statHandle("# client CHK content deleted (kb)");
  stat_cs_delete_3hash_count 
    = statHandle("# client 3HASH search results deleted (kb)");
  stat_cs_unindex_block_count 
    = statHandle("# client file unindex requests received");
  stat_cs_unindex_file_count 
    = statHandle("# file unindex requests received");
  stat_cs_unindex_super_count 
    = statHandle("# super query unindex requests received");
  stat_cs_insert_sblock_count
    = statHandle("# client SBlock insert requests received");
  stat_cs_nsquery_count
    = statHandle("# client namespace queries received");
  stat_p2p_nsquery_count
    = statHandle("# peer namespace queries received");
  stat_p2p_sblock_replies
    = statHandle("# SBlocks received");
  return OK;
}

/**
 * Handle query for content. Depending on how we like the sender,
 * lookup, forward or even indirect.
 **/
int handleQUERY(HostIdentity * sender,
		p2p_HEADER * msg) {
  QUERY_POLICY qp;
  AFS_p2p_QUERY * qmsg;
#if DEBUG_HANDLER
  HexName hex;
  HexName hex2;
#endif
  int queries;
  int ttl;
  unsigned int prio;
  double preference;
  int count;
      

  queries = (ntohs(msg->size) - sizeof(AFS_p2p_QUERY)) / sizeof(HashCode160);
  if ( (queries <= 0) || 
       (ntohs(msg->size) != sizeof(AFS_p2p_QUERY) + queries * sizeof(HashCode160)) ) {
    LOG(LOG_WARNING,
	"WARNING: query received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_query_count, 1);
  qmsg = (AFS_p2p_QUERY*) msg;

#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&qmsg->queries[0],
		 &hex));
  IFLOG(LOG_EVERYTHING,
	hash2hex(&sender->hashPubKey,
		 &hex2));
  LOG(LOG_EVERYTHING,
      "EVERYTHING: received query %s (%d) TTL %d PR %u from %s\n",
      &hex,
      queries,
      ntohl(qmsg->ttl),
      ntohl(qmsg->priority),
      &hex2);
#endif

  /* decrement ttl (always) */
  ttl = ntohl(qmsg->ttl);
#if DEBUG_HANDLER
  LOG(LOG_DEBUG,
      "DEBUG: received query for %s with ttl %d\n",
      &hex,
      ttl);
#endif
  if (ttl < 0) {
    ttl = ttl - 2*TTL_DECREMENT - randomi(TTL_DECREMENT);
    if (ttl > 0)
      return OK; /* just abort */
  } else
    ttl = ttl - 2*TTL_DECREMENT - randomi(TTL_DECREMENT);
  qp = evaluateQuery(sender,
		     ntohl(qmsg->priority));  
  if ((qp & QUERY_DROPMASK) == 0)
    return OK; /* straight drop. */

  preference = (double) (qp & QUERY_PRIORITY_BITMASK);
  if (preference < QUERY_BANDWIDTH_VALUE)
    preference = QUERY_BANDWIDTH_VALUE;
  coreAPI->preferTrafficFrom(sender,
			     preference);

  /* adjust priority */
  prio = ntohl(qmsg->priority);
  if ( (qp & QUERY_PRIORITY_BITMASK) < prio) {
    prio = qp & QUERY_PRIORITY_BITMASK;
    qmsg->priority = htonl(prio);
  }  
  count = (ntohs(qmsg->header.size)-sizeof(AFS_p2p_QUERY)) / sizeof(HashCode160);
  prio = prio / count; /* effective priority for ttl */
  
  /* adjust TTL */
  if ( (ttl > 0) &&
       (ttl > (int)prio*1000) ) 
    ttl = (int) prio*1000; /* bound! */
  qmsg->ttl = htonl(ttl);

  execQuery(qp, qmsg, NULL);
  return OK;
}
 
/**
 * Receive content, do something with it!  There are 3 basic
 * possiblilities. Either our node did the request and we should send
 * the result to a client via TCP, or the content was requested by
 * another node and we forwarded the request (and thus we now have to
 * fwd the reply) or 3rd somebody just send us some content we did NOT
 * ask for - and we can choose to store it or just discard it.
 **/
int handleCHK_CONTENT(HostIdentity * sender, 
		      p2p_HEADER * msg) {
  int prio;
  HashCode160 queryHash;
  ContentIndex ce;
  AFS_p2p_CHK_RESULT * cmsg;
  int ret;
  int dupe;
  double preference;

  if (ntohs(msg->size) != sizeof(AFS_p2p_CHK_RESULT)) {
    LOG(LOG_WARNING,
	"WARNING: CHK content message received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_chk_replies, 1);
  cmsg = (AFS_p2p_CHK_RESULT*) msg;
  hash(&cmsg->result,
       CONTENT_SIZE,
       &queryHash);
  prio = useContent(sender,
		    &queryHash,
		    msg);
  if (sender == NULL) /* no migration, this is already content
			 from the local node */
    return OK;  
  preference = (double) prio;
  prio = evaluateContent(&queryHash,
			 prio);
  if (prio != SYSERR)
    preference += (double) prio;
  if (preference < CONTENT_BANDWIDTH_VALUE)
    preference = CONTENT_BANDWIDTH_VALUE;
  coreAPI->preferTrafficFrom(sender,
			     preference);

  if (prio == SYSERR)
    return OK; /* straight drop */
  memcpy(&ce.hash,
	 &queryHash,
	 sizeof(HashCode160));
  ce.importance    = htonl(prio);
  ce.type          = htons(LOOKUP_TYPE_CHK);
  ce.fileNameIndex = 0;
  ce.fileOffset    = 0;
  ret = insertContent(&ce, 
		      sizeof(CONTENT_Block),
		      &cmsg->result, 
		      sender,
		      &dupe);
  if ( (ret == OK) &&
       (dupe == NO) )
    addToBloomfilter(singleBloomFilter,
		     &queryHash);
  return OK;
}

/**
 * Receive content, do something with it!  There are 3 basic
 * possiblilities. Either our node did the request and we should send
 * the result to a client via TCP, or the content was requested by
 * another node and we forwarded the request (and thus we now have to
 * fwd the reply) or 3rd somebody just send us some content we did NOT
 * ask for - and we can choose to store it or just discard it.
 **/
int handle3HASH_CONTENT(HostIdentity * sender, 
			p2p_HEADER * msg) {
  int prio;
  AFS_p2p_3HASH_RESULT * cmsg;
  HashCode160 tripleHash;
  ContentIndex ce;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ret;
  int dupe;
  double preference;

  if (ntohs(msg->size) != sizeof(AFS_p2p_3HASH_RESULT)) {
    LOG(LOG_WARNING,
	"WARNING: content message received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_3hash_replies, 1);
  cmsg = (AFS_p2p_3HASH_RESULT*) msg;
  hash(&cmsg->hash,
       sizeof(HashCode160),
       &tripleHash);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&tripleHash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received 3HASH search result for %s from peer\n",
      &hex);
#endif
  prio = useContent(sender,
		    &tripleHash,
		    msg);
  if (sender == NULL) { /* no migration, this is already content
			   from the local node */
#if DEBUG_HANDLER
    LOG(LOG_DEBUG,
	"DEBUG: content migration not needed, content is local\n");
#endif
    return OK;  
  }
  preference = (double) prio;
#if DEBUG_HANDLER
  LOG(LOG_DEBUG,
      "DEBUG: content migration with preference %d\n",
      prio);
#endif
  prio = evaluateContent(&tripleHash,
			 prio);
  if (prio != SYSERR)
    preference += (double) prio;
  if (preference < CONTENT_BANDWIDTH_VALUE)
    preference = CONTENT_BANDWIDTH_VALUE;
  coreAPI->preferTrafficFrom(sender,
			     preference);


  if (prio == SYSERR) {
#if DEBUG_HANDLER
    LOG(LOG_DEBUG,
	"DEBUG: content not important enough, not replicated\n");
#endif
    return OK; /* straight drop */
  } 
#if DEBUG_HANDLER
  else
    LOG(LOG_DEBUG,
	"DEBUG: content replicated with total preference %d\n",
	prio);
#endif
  memcpy(&ce.hash,
	 &cmsg->hash,
	 sizeof(HashCode160));
  ce.importance    = htonl(prio);
  ce.type          = htons(LOOKUP_TYPE_3HASH);
  ce.fileNameIndex = 0;
  ce.fileOffset    = 0;
  
  ret = insertContent(&ce, 
		      sizeof(CONTENT_Block),
             	      &cmsg->result, 
                      sender,
		      &dupe);
  if ( (ret == OK) &&
       (dupe == NO) )
      addToBloomfilter(singleBloomFilter,
	               &tripleHash);
  return OK;
}

/* *********************** CS handlers ***************** */

/**
 * Process a query from the client. Forwards to the network.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/ 
int csHandleRequestQuery(ClientHandle sock,
			 AFS_CS_QUERY * queryRequest) {
  QUERY_POLICY qp = QUERY_ANSWER|QUERY_FORWARD|QUERY_INDIRECT|QUERY_PRIORITY_BITMASK; 
  AFS_p2p_QUERY * msg;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int queries;
  int ttl;
  int ret;

  queries = (ntohs(queryRequest->header.size) - sizeof(AFS_CS_QUERY)) / sizeof(HashCode160);
  if ( (queries <= 0) ||
       (ntohs(queryRequest->header.size) != 
	sizeof(AFS_CS_QUERY) + queries * sizeof(HashCode160)) ) {
    LOG(LOG_WARNING,
	"WARNING: received malformed query from client\n");
    coreAPI->sendTCPResultToClient(sock, SYSERR);
    return SYSERR;
  }
  statChange(stat_cs_query_count, 1);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&queryRequest->queries[0], 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: received %d queries (%s) with ttl %d and priority %u.\n",
      queries,
      &hex,
      ntohl(queryRequest->ttl),
      ntohl(queryRequest->priority));
#endif
  msg = MALLOC(sizeof(AFS_p2p_QUERY)+queries * sizeof(HashCode160));
  msg->header.size 
    = htons(sizeof(AFS_p2p_QUERY)+queries * sizeof(HashCode160));
  msg->header.requestType 
    = htons(AFS_p2p_PROTO_QUERY);
  memcpy(&msg->queries[0],
	 &queryRequest->queries[0],
	 sizeof(HashCode160) * queries);
  msg->priority 
    = queryRequest->priority; /* no htonl here: is already in network byte order! */
  /* adjust TTL */
  ttl = ntohl(queryRequest->ttl);
  if ( (ttl > 0) &&
       (ttl > (int)ntohl(msg->priority)/queries*1000) ) 
    ttl = (int) ntohl(msg->priority)/queries*1000; /* bound! */
  msg->ttl = htonl(ttl);
  memcpy(&msg->returnTo,
	 coreAPI->myIdentity,
	 sizeof(HostIdentity));
  ret = execQuery(qp, msg, sock);   
  
  FREE(msg);
  return coreAPI->sendTCPResultToClient(sock, ret);
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestInsertCHK(ClientHandle sock,
			     AFS_CS_INSERT_CHK * insertRequest) {
  ContentIndex entry;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ret;
  int dupe;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_CHK)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed CHK insert request from client\n");
    return SYSERR;
  } 
  statChange(stat_cs_insert_chk_count, 1);
  hash(&insertRequest->content,
       sizeof(CONTENT_Block),
       &entry.hash);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&entry.hash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received CHK insert request for block %s\n",
      &hex);
#endif
  entry.type
    = htons(LOOKUP_TYPE_CHK);
  entry.importance
    = insertRequest->importance; /* both are in network byte order! */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */

  ret = insertContent(&entry,
     	              sizeof(CONTENT_Block),
		      &insertRequest->content,
		      NULL,
		      &dupe);
  if ( (ret == OK) &&
       (dupe == NO) )
      addToBloomfilter(singleBloomFilter,
	               &entry.hash);
  return coreAPI->sendTCPResultToClient(sock, ret);
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestInsert3HASH(ClientHandle sock,
			       AFS_CS_INSERT_3HASH * insertRequest) {
  ContentIndex entry;
  HashCode160 tripleHash;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int dupe;
  int ret;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_3HASH)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed 3HASH insert request from client\n");
    return SYSERR;
  }
  statChange(stat_cs_insert_3hash_count, 1);
  memcpy(&entry.hash,
	 &insertRequest->doubleHash,
	 sizeof(HashCode160));
  hash(&insertRequest->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&tripleHash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received 3HASH insert request for %s from client\n",
      &hex);
#endif
  entry.type
    = htons(LOOKUP_TYPE_3HASH);
  entry.importance
    = insertRequest->importance; /* both are in network byte order! */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */
  ret = insertContent(&entry,
		      sizeof(CONTENT_Block),
		      &insertRequest->content,
	   	      NULL,
		      &dupe);
  if ( (ret == OK) &&
       (dupe == NO) )
    addToBloomfilter(singleBloomFilter,
		     &tripleHash);
		     
  return coreAPI->sendTCPResultToClient(sock, ret);
}

/**
 * Process a request to index content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestIndexBlock(ClientHandle sock,
			      AFS_CS_INDEX_BLOCK * indexingRequest) {
  int dupe;
  
  if (ntohs(indexingRequest->header.size) != 
      sizeof(AFS_CS_INDEX_BLOCK)) {
    LOG(LOG_WARNING, 
	"WARNING: block indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_block_count, 1);
  return coreAPI->sendTCPResultToClient(sock,
		       insertContent(&indexingRequest->contentIndex,
				     0, 
				     NULL, 
				     NULL, 
				     &dupe));
}

/**
 * Process a query to list a file as on-demand encoded from the client.
 *
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestIndexFile(ClientHandle sock,
			     AFS_CS_INDEX_FILE * listFileRequest) {
  if (ntohs(listFileRequest->header.size) != 
      sizeof(AFS_CS_INDEX_FILE)) {
    LOG(LOG_WARNING, 
	"WARNING: file indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_file_count, 1);
  listFileRequest->filename[1023] = 0; /* ensure strlen terminates! */
  return coreAPI->sendTCPResultToClient(sock, 
		       appendFilename(listFileRequest->filename));
}

/**
 * Process a client request to extend our super-query bloom
 * filter.
 **/
int csHandleRequestIndexSuper(ClientHandle sock,
			      AFS_CS_INDEX_SUPER * superIndexRequest) {
  ContentIndex entry;
  int dupe;

  if (ntohs(superIndexRequest->header.size) != 
      sizeof(AFS_CS_INDEX_SUPER)) {
    LOG(LOG_WARNING, 
	"WARNING: super-hash indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_super_count, 1);
  addToBloomfilter(superBloomFilter,
		   &superIndexRequest->superHash);
  entry.type
    = htons(LOOKUP_TYPE_SUPER);
  entry.importance
    = superIndexRequest->importance; /* both are in network byte order */ 
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */
  memcpy(&entry.hash,
	 &superIndexRequest->superHash,
	 sizeof(HashCode160));
  return coreAPI->sendTCPResultToClient(sock, 
		       insertContent(&entry,
				     0, 
				     NULL, 
				     NULL, 
				     &dupe));
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestDeleteCHK(ClientHandle sock,
			     AFS_CS_INSERT_CHK * insertRequest) {
  HashCode160 hc;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ret;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_CHK)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed CHK remove request from client\n");
    return SYSERR;
  } 
  statChange(stat_cs_delete_chk_count, 1);
  hash(&insertRequest->content,
       sizeof(CONTENT_Block),
       &hc);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&hc,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received CHK remove request for block %s\n",
      &hex);
#endif
  ret = removeContent(&hc,
                      -1);
  if (ret == OK)
    if (YES == testBloomfilter(singleBloomFilter,
			       &hc))
      delFromBloomfilter(singleBloomFilter,
			 &hc);
  return coreAPI->sendTCPResultToClient(sock, ret);
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestDelete3HASH(ClientHandle sock,
			       AFS_CS_INSERT_3HASH * insertRequest) {
  HashCode160 tripleHash;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ret;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_3HASH)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed 3HASH delete request from client\n");
    return SYSERR;
  }
  statChange(stat_cs_delete_3hash_count, 1);
  hash(&insertRequest->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&tripleHash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received 3HASH delete request for %s from client\n",
      &hex);
#endif
  ret = removeContent(&tripleHash,
                      -1);
  if (ret == OK)     
    delFromBloomfilter(singleBloomFilter,
		       &tripleHash);
		     
  return coreAPI->sendTCPResultToClient(sock, ret);
}

/**
 * Process a request to index content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestUnindexBlock(ClientHandle sock,
				AFS_CS_INDEX_BLOCK * indexingRequest) {
  if (ntohs(indexingRequest->header.size) != 
      sizeof(AFS_CS_INDEX_BLOCK)) {
    LOG(LOG_WARNING, 
	"WARNING: block unindexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_unindex_block_count, 1);
  return coreAPI->sendTCPResultToClient(sock,
		       removeContent(&indexingRequest->contentIndex.hash,
		                     -1));
}

/**
 * Callback used to select the file in the fileindex
 * that is to be removed.
 **/
static int removeMatch(char * fn,
		       int i,
		       char * search) {
  if (strcmp(fn, search) == 0)
    return SYSERR;
  else
    return OK;     
}

/**
 * Process a query to list a file as on-demand encoded from the client.
 *
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestUnindexFile(ClientHandle sock,
			       AFS_CS_INDEX_FILE * listFileRequest) {
  int idx;

  if (ntohs(listFileRequest->header.size) != 
      sizeof(AFS_CS_INDEX_FILE)) {
    LOG(LOG_WARNING, 
	"WARNING: file unindexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_file_count, 1);
  listFileRequest->filename[1023] = 0; /* ensure strlen terminates! */

  idx = appendFilename(listFileRequest->filename);
  forEachIndexedFile((IndexedFileNameCallback)&removeMatch,
		     listFileRequest->filename);
  return coreAPI->sendTCPResultToClient(sock, 
		       idx);
}

/**
 * Process a client request to extend our super-query bloom
 * filter.
 **/
int csHandleRequestUnindexSuper(ClientHandle sock,
				AFS_CS_INDEX_SUPER * superIndexRequest) {
  if (ntohs(superIndexRequest->header.size) != 
      sizeof(AFS_CS_INDEX_SUPER)) {
    LOG(LOG_WARNING, 
	"WARNING: super-hash unindexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_unindex_super_count, 1);
  delFromBloomfilter(superBloomFilter,
		     &superIndexRequest->superHash);
  return coreAPI->sendTCPResultToClient(sock, 
		       removeContent(&superIndexRequest->superHash,
		       -1));
}

/* *************************** SBlock stuff ***************************** */

int csHandleRequestInsertSBlock(ClientHandle sock,
				AFS_CS_INSERT_SBLOCK * insertRequest) {
  ContentIndex entry;
#if DEBUG_HANDLER
  HexName hex1;
  HexName hex2;
  HashCode160 ns;
#endif
  int dupe;
  int ret;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_SBLOCK)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed SBLOCK insert request from client\n");
    return SYSERR;
  }
  statChange(stat_cs_insert_sblock_count, 1);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&insertRequest->content.identifier,
		 &hex1);
	hash(&insertRequest->content.subspace,
	     sizeof(PublicKey),
	     &ns);
	hash2hex(&ns,
		 &hex2));
  LOG(LOG_DEBUG,
      "DEBUG: received SBlock for namespace %s with routing ID %s.\n",
      &hex2,
      &hex1);
#endif
  entry.type
    = htons(LOOKUP_TYPE_SBLOCK);
  entry.importance
    = insertRequest->importance; /* both are in network byte order! */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */
  memcpy(&entry.hash,
	 &insertRequest->content.identifier,
	 sizeof(HashCode160));
  dupe = NO;
  ret = insertContent(&entry,
		      sizeof(CONTENT_Block),
		      &insertRequest->content,
	   	      NULL,
		      &dupe);
#if DEBUG_HANDLER
  LOG(LOG_DEBUG,
      "DEBUG: received SBlock insert is dupe: %s (insert %s)\n",
      dupe == NO ? "NO" : "YES",
      ret == SYSERR ? "SYSERR" : "OK");
#endif
  if ( (ret == OK) &&
       (dupe == NO) )
    addToBloomfilter(singleBloomFilter,
		     &insertRequest->content.identifier);
  return coreAPI->sendTCPResultToClient(sock, ret);
}

int csHandleRequestNSQuery(ClientHandle sock,
			   AFS_CS_NSQUERY * queryRequest) {
  QUERY_POLICY qp = QUERY_ANSWER|QUERY_FORWARD|QUERY_INDIRECT|QUERY_PRIORITY_BITMASK; 
  AFS_p2p_NSQUERY * msg;
#if DEBUG_HANDLER
  HexName hex1;
  HexName hex2;
#endif

  if (ntohs(queryRequest->header.size) != 
      sizeof(AFS_CS_NSQUERY)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed NS query from client\n");
    return SYSERR;
  }
  statChange(stat_cs_nsquery_count, 1);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&queryRequest->identifier, 
		 &hex1));
  IFLOG(LOG_DEBUG,
	hash2hex(&queryRequest->namespace, 
		 &hex2));
  LOG(LOG_DEBUG, 
      "DEBUG: received NS query (%s/%s) with ttl %d and priority %u.\n",
      &hex2,
      &hex1,
      ntohl(queryRequest->ttl),
      ntohl(queryRequest->priority));
#endif
  msg = MALLOC(sizeof(AFS_p2p_NSQUERY));
  msg->hdr.header.size 
    = htons(sizeof(AFS_p2p_NSQUERY));
  msg->hdr.header.requestType 
    = htons(AFS_p2p_PROTO_NSQUERY);
  msg->hdr.priority 
    = queryRequest->priority; /* no htonl here: is already in network byte order! */
  msg->hdr.ttl 
    = queryRequest->ttl; /* no htonl here: is already in network byte order! */
  memcpy(&msg->identifier,
	 &queryRequest->identifier,
	 sizeof(HashCode160));
  memcpy(&msg->namespace,
	 &queryRequest->namespace,
	 sizeof(HashCode160));
  memcpy(&msg->hdr.returnTo,
	 coreAPI->myIdentity,
	 sizeof(HostIdentity));
  execQuery(qp, &msg->hdr, sock);   
  FREE(msg);
  return OK;
}

int handleNSQUERY(HostIdentity * sender,
		  p2p_HEADER * msg) {
  QUERY_POLICY qp;
  AFS_p2p_NSQUERY * qmsg;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ttl;
  double preference;
  
  if (ntohs(msg->size) != sizeof(AFS_p2p_NSQUERY)) {
    LOG(LOG_WARNING,
	"WARNING: nsquery received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_nsquery_count, 1);
  qmsg = (AFS_p2p_NSQUERY*) msg;
  /* decrement ttl */
  ttl = ntohl(qmsg->hdr.ttl);
#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&qmsg->identifier,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received NS query for %s with ttl %d\n",
      &hex,
      ttl);
#endif
  if (ttl < 0) {
    ttl = ttl - 2*TTL_DECREMENT - randomi(TTL_DECREMENT);
    if (ttl > 0)
      return OK; /* just abort */
  } else
    ttl = ttl - 2*TTL_DECREMENT - randomi(TTL_DECREMENT);
  qmsg->hdr.ttl = htonl(ttl);
  qp = evaluateQuery(sender,
		     ntohl(qmsg->hdr.priority));  
  if ((qp & QUERY_DROPMASK) == 0)
    return OK; /* straight drop. */

  preference = (double) (qp & QUERY_PRIORITY_BITMASK);
  if (preference < QUERY_BANDWIDTH_VALUE)
    preference = QUERY_BANDWIDTH_VALUE;
  coreAPI->preferTrafficFrom(sender,
			     preference);

  execQuery(qp, &qmsg->hdr, NULL);
  return OK;
}


int handleSBLOCK_CONTENT(HostIdentity * sender, 
			 p2p_HEADER * msg) {
  int prio;
  AFS_p2p_SBLOCK_RESULT * cmsg;
  ContentIndex ce;
#if DEBUG_HANDLER
  HexName hex;
#endif
  int ret;
  int dupe;
  double preference;

  if (ntohs(msg->size) != sizeof(AFS_p2p_SBLOCK_RESULT)) {
    LOG(LOG_WARNING,
	"WARNING: signed content message received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_sblock_replies, 1);
  cmsg = (AFS_p2p_SBLOCK_RESULT*) msg;

  if (OK != verifySBlock(&cmsg->result))
    return SYSERR;

#if DEBUG_HANDLER
  IFLOG(LOG_DEBUG,
	hash2hex(&cmsg->result.identifier,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received SBLOCK search result for %s from peer\n",
      &hex);
#endif
  prio = useContent(sender,
		    &cmsg->result.identifier,
		    msg);
  if (sender == NULL) { /* no migration, this is already content
			   from the local node */
#if DEBUG_HANDLER
    LOG(LOG_DEBUG,
	"DEBUG: content migration not needed, content is local\n");
#endif
    return OK;  
  }
#if DEBUG_HANDLER
  else
    LOG(LOG_DEBUG,
	"DEBUG: content migration with preference %d\n",
	prio);
#endif
  preference = (double) prio;
  prio = evaluateContent(&cmsg->result.identifier,
			 prio);
  if (prio == SYSERR) {
#if DEBUG_HANDLER
    LOG(LOG_DEBUG,
	"DEBUG: content not important enough, not replicated\n");
#endif
    return OK; /* straight drop */
  } 
#if DEBUG_HANDLER
  else
    LOG(LOG_DEBUG,
	"DEBUG: content replicated with total preference %d\n",
	prio);
#endif
  if (prio != SYSERR)
    preference += (double) prio;
  if (preference < CONTENT_BANDWIDTH_VALUE)
    preference = CONTENT_BANDWIDTH_VALUE;
  coreAPI->preferTrafficFrom(sender,
			     preference);
  memcpy(&ce.hash,
	 &cmsg->result.identifier,
	 sizeof(HashCode160));
  ce.importance    = htonl(prio);
  ce.type          = htons(LOOKUP_TYPE_SBLOCK);
  ce.fileNameIndex = 0;
  ce.fileOffset    = 0;
  
  ret = insertContent(&ce, 
		      sizeof(CONTENT_Block),
             	      &cmsg->result, 
                      sender,
		      &dupe);
  if ( (ret == OK) &&
       (dupe == NO) )
      addToBloomfilter(singleBloomFilter,
	               &cmsg->result.identifier);
  return OK;
}


/* end of handler.c */
